"""
dataclasses module for VMDR module - contains the dataclasses for the VMDR module.
"""

from .software import Software
from .vendor_reference import VendorReference
from .kb_entry import KBEntry
from .bugtraq import Bugtraq
from .cve import CVEID
from .threat_intel import ThreatIntel
from .compliance import Compliance
from .tag import Tag, CloudTag
from .detection import Detection
from .qds_factor import QDSFactor
from .qds import QDS
from .asset_group import AssetGroup
from .vmscan import VMScan
from .scanner_appliance import ScannerAppliance
from .searchlist import StaticSearchList
from .report import VMDRReport, VMDRScheduledReport
from .report_template import ReportTemplate
from .user import User

from .lists import BaseList



"""
static_searchlists.py - Contains functions to interact with Static Searchlists in VMDR.
"""

from typing import *

from ..exceptions import *
from ..auth import BasicAuth
from ..base import call_api, xml_parser
from .data_classes import BaseList
from .data_classes import StaticSearchList


def get_static_searchlists(
    auth: BasicAuth, ids: str = None
) -> BaseList[StaticSearchList]:
    """
    Get a list of Static Searchlists in VMDR.

    Parameters:
        auth (BasicAuth): The authentication to use.
        ids (str): A comma-separated list of Static Searchlist IDs to retrieve.

    Returns:
        BaseList[StaticSearchlist]: A list of Static Searchlists.
    """
    responses = BaseList()
    params = {}
    params["action"] = "list"
    if ids:
        params["ids"] = ids

    resp = call_api(
        auth=auth,
        module="vmdr",
        endpoint="get_static_searchlists",
        params=params,
        headers={"X-Requested-With": "qualysdk SDK"},
    )

    searchlists = xml_parser(resp.text)

    if "STATIC_LISTS" not in searchlists["STATIC_SEARCH_LIST_OUTPUT"]["RESPONSE"]:
        print("No Static Searchlists found.")
        return responses

    # If there is only one searchlist, it will not be in a list.
    if isinstance(
        searchlists["STATIC_SEARCH_LIST_OUTPUT"]["RESPONSE"]["STATIC_LISTS"][
            "STATIC_LIST"
        ],
        dict,
    ):
        searchlists["STATIC_SEARCH_LIST_OUTPUT"]["RESPONSE"]["STATIC_LISTS"][
            "STATIC_LIST"
        ] = [
            searchlists["STATIC_SEARCH_LIST_OUTPUT"]["RESPONSE"]["STATIC_LISTS"][
                "STATIC_LIST"
            ]
        ]

    for searchlist in searchlists["STATIC_SEARCH_LIST_OUTPUT"]["RESPONSE"][
        "STATIC_LISTS"
    ]["STATIC_LIST"]:
        responses.append(StaticSearchList(**searchlist))

    return responses



"""
VMDR (Vulnerability Management) module

This module contains ways to interact with the Qualys VMDR API. Valid endpoints are defined in the CALL_SCHEMA dictionary.

VMDR QQL Syntax help: https://qualysguard.qg2.apps.qualys.com/portal-help/en/vm/search/how_to_search.htm
"""

from .query_kb import query_kb
from .get_host_list import get_host_list
from .get_host_list_detections import get_hld
from .ips import get_ip_list, add_ips, update_ips
from .assetgroups import get_ag_list, add_ag, edit_ag, delete_ag
from .vmscans import (
    get_scan_list,
    launch_scan,
    pause_scan,
    resume_scan,
    cancel_scan,
    fetch_scan,
    delete_scan,
)
from .scanner_appliances import get_scanner_list
from .searchlists import get_static_searchlists
from .reports import (
    get_report_list,
    get_template_list,
    launch_report,
    cancel_report,
    fetch_report,
    delete_report,
    get_scheduled_report_list,
    launch_scheduled_report,
)

from .users import get_user_list, add_user, edit_user



"""
asset_group.py - contains the AssetGroup dataclass for the Qualys VMDR module.
"""

from dataclasses import dataclass, field, asdict
from typing import *
from datetime import datetime
from ipaddress import (
    IPv4Address,
    IPv6Address,
    IPv4Network,
    IPv6Network,
)

from .ip_converters import *
from .hosts import VMDRID
from .lists.base_list import BaseList


@dataclass(order=True)
class AssetGroup:
    """
    AssetGroup - represents a single asset group in Qualys.
    """

    ID: int = field(metadata={"description": "The ID of the asset group."})
    TITLE: Optional[str] = field(
        metadata={"description": "The title of the asset group."}, default=""
    )
    # NOTE: OWNER_ID contains either the raw XML's OWNER_ID or OWNER_USER_ID depending on value passed to get_ag_list(attributes=<>).
    OWNER_ID: Optional[int] = field(
        metadata={"description": "The ID of the owner of the asset group."},
        default=None,
    )
    OWNER_USER_ID: Optional[int] = field(
        metadata={"description": "The ID of the owner user of the asset group."},
        default=None,
    )
    UNIT_ID: Optional[int] = field(
        metadata={"description": "The ID of the unit of the asset group."}, default=None
    )
    LAST_UPDATE: Optional[Union[str, datetime]] = field(
        metadata={"description": "The datetime the asset group was last updated."},
        default=None,
    )
    NETWORK_ID: Optional[int] = field(
        metadata={
            "description": "The ID of the network of the asset group, if enabled in Qualys."
        },
        default=None,
    )
    IP_SET: Optional[
        BaseList[Union[IPv4Address, IPv6Address, IPv4Network, IPv6Network]]
    ] = field(
        metadata={"description": "The IP set of the asset group."},
        default=None,
    )
    BUSINESS_IMPACT: Optional[str] = field(
        metadata={"description": "The business impact of the asset group."},
        default=None,
    )
    DEFAULT_APPLIANCE_ID: Optional[int] = field(
        metadata={"description": "The default appliance ID of the asset group."},
        default=None,
    )
    APPLIANCE_IDS: Optional[BaseList[int]] = field(
        metadata={"description": "The appliance IDs of the asset group."},
        default=None,
    )
    DNS_LIST: Optional[BaseList[str]] = field(
        metadata={"description": "The DNS list of the asset group."},
        default=None,
    )
    NETBIOS_LIST: Optional[BaseList[str]] = field(
        metadata={"description": "The NetBIOS list of the asset group."},
        default=None,
    )
    HOST_IDS: Optional[BaseList[VMDRID]] = field(
        metadata={
            "description": "The host IDs of the asset group. BaseList of VMDRID objects."
        },
        default=None,
    )
    ASSIGNED_USER_IDS: Optional[BaseList[int]] = field(
        metadata={"description": "The assigned user IDs of the asset group."},
        default=None,
    )
    ASSIGNED_UNIT_IDS: Optional[BaseList[int]] = field(
        metadata={"description": "The assigned unit IDs of the asset group."},
        default=None,
    )
    OWNER_USER_NAME: Optional[str] = field(
        metadata={"description": "The owner user name of the asset group."},
        default=None,
    )
    CVSS_ENVIRO_CDP: Optional[str] = field(
        metadata={"description": "The CVSS environmental CDP of the asset group."},
        default=None,
    )
    CVSS_ENVIRO_TD: Optional[str] = field(
        metadata={"description": "The CVSS environmental TD of the asset group."},
        default=None,
    )
    CVSS_ENVIRO_CR: Optional[str] = field(
        metadata={"description": "The CVSS environmental CR of the asset group."},
        default=None,
    )
    CVSS_ENVIRO_IR: Optional[str] = field(
        metadata={"description": "The CVSS environmental IR of the asset group."},
        default=None,
    )
    CVSS_ENVIRO_AR: Optional[str] = field(
        metadata={"description": "The CVSS environmental AR of the asset group."},
        default=None,
    )
    EC2_IDS: Optional[BaseList[str]] = field(
        metadata={"description": "The EC2 IDs of the asset group."},
        default=None,
    )
    COMMENTS: Optional[BaseList[str]] = field(
        metadata={"description": "The comments of the asset group."},
        default=None,
    )
    DOMAIN_LIST: Optional[BaseList[str]] = field(
        metadata={"description": "The domain list of the asset group."},
        default=None,
    )

    def __post_init__(self):
        # Thanks Qualys for the inconsistency in naming conventions:
        if self.OWNER_USER_ID:
            self.OWNER_ID = self.OWNER_USER_ID
            del self.OWNER_USER_ID

        # Do data conversions:
        INT_FIELDS = ["ID", "OWNER_ID", "UNIT_ID", "NETWORK_ID", "DEFAULT_APPLIANCE_ID"]
        DT_FIELDS = ["LAST_UPDATE"]
        INT_LISTS = ["APPLIANCE_IDS", "ASSIGNED_USER_IDS", "ASSIGNED_UNIT_IDS"]
        STR_LISTS = ["DNS_LIST", "NETBIOS_LIST", "EC2_IDS", "COMMENTS", "DOMAIN_LIST"]

        for field in STR_LISTS:
            if getattr(self, field):
                setattr(self, field, BaseList(getattr(self, field).split(",")))

        for field in INT_LISTS:
            if getattr(self, field):
                if isinstance(getattr(self, field), str):
                    setattr(
                        self,
                        field,
                        BaseList([int(x) for x in getattr(self, field).split(",")]),
                    )
                else:
                    setattr(
                        self, field, BaseList([int(x) for x in getattr(self, field)])
                    )

        for field in DT_FIELDS:
            if getattr(self, field):
                setattr(self, field, datetime.fromisoformat(getattr(self, field)))

        for field in INT_FIELDS:
            if getattr(self, field):
                setattr(self, field, int(getattr(self, field)))

        # Convert IP_SET to BaseList of ipaddress.* objs.:
        if self.IP_SET:
            final_ip_set = BaseList()
            if self.IP_SET.get("IP_RANGE"):
                if isinstance(self.IP_SET.get("IP_RANGE"), str):
                    # We can use single_range here because it's a single IP range.
                    final_ip_set.extend([single_range(self.IP_SET.get("IP_RANGE"))])
                else:
                    # We can use convert_ranges here because it's a list of IP ranges.
                    final_ip_set.extend(convert_ranges(self.IP_SET.get("IP_RANGE")))

            if self.IP_SET.get("IP"):
                if isinstance(self.IP_SET.get("IP"), str):
                    # We can use single_ip here because it's a single IP.
                    final_ip_set.extend([single_ip(self.IP_SET.get("IP"))])
                else:
                    # We can use convert_ips here because it's a list of IPs.
                    final_ip_set.extend(convert_ips(self.IP_SET.get("IP")))

            self.IP_SET = final_ip_set

        # Convert HOST_IDS to BaseList of VMDRID objs.:
        if self.HOST_IDS:
            final_host_ids = BaseList()
            if isinstance(self.HOST_IDS, str):
                final_host_ids.extend(
                    [
                        VMDRID(ID=host_id, TYPE="host")
                        for host_id in self.HOST_IDS.split(",")
                    ]
                )
            else:
                final_host_ids.extend(
                    [VMDRID(ID=host_id, TYPE="host") for host_id in self.HOST_IDS]
                )

            self.HOST_IDS = final_host_ids

    def __str__(self):
        return self.TITLE

    def __contains__(self, item):
        return (
            item in self.ID
            or item in self.TITLE
            or item in self.OWNER_ID
            or item in self.UNIT_ID
            or item in self.NETWORK_ID
            or item in self.IP_SET
        )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def copy(self):
        return AssetGroup(
            ID=self.ID,
            TITLE=self.TITLE,
            OWNER_ID=self.OWNER_ID,
            UNIT_ID=self.UNIT_ID,
            NETWORK_ID=self.NETWORK_ID,
            IP_SET=self.IP_SET,
        )

    def is_id(self, id: int):
        return self.ID == id

    def is_title(self, title: str):
        return self.TITLE == title

    def is_owner_id(self, owner_id: int):
        return self.OWNER_ID == owner_id

    def is_unit_id(self, unit_id: int):
        return self.UNIT_ID == unit_id

    def is_network_id(self, network_id: int):
        return self.NETWORK_ID == network_id

    def contains_ip(
        self, ip: Union[IPv4Address, IPv6Address, IPv4Network, IPv6Network]
    ):
        return ip in self.IP_SET

    def add_ip(self, ip: Union[IPv4Address, IPv6Address, IPv4Network, IPv6Network]):
        self.IP_SET.append(ip)

    def remove_ip(self, ip: Union[IPv4Address, IPv6Address, IPv4Network, IPv6Network]):
        self.IP_SET.remove(ip)

    def valid_values(self):
        """
        Return a dictionary of non-None attributes.
        """
        return {
            key: value
            for key, value in self.to_dict().items()
            if value is not None and value != [] and value != ""
        }

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    def to_dict(self):
        """
        to_dict - convert the AssetGroup object to a dictionary.

        This function is used to convert the AssetGroup object to a dictionary.
        """
        return asdict(self)

    def __dict__(self):
        return asdict(self)

    @classmethod
    def from_dict(cls, data: dict):
        """
        from_dict - create an AssetGroup object from a dictionary.

        This function is used to create an AssetGroup object from a dictionary.
        """
        required_keys = {"ID", "TITLE"}
        if not required_keys.issubset(data.keys()):
            raise ValueError(
                f"Dictionary must contain the following keys: {required_keys}"
            )

        return cls(**data)

    def to_sql(self) -> dict:
        """
        Prepare the dataclass for insertion into a SQL database
        by converting it to a dictionary with appropriate list
        and datetime conversions.
        """

        # TODO: NOTE: INSTEAD OF ASDICT, MAYBE TRY ACCESSING THE ATTRIBUTES DIRECTLY, MODIFYING THEM, AND THEN DOING ASDICT???
        # NOTE: ABOVE WORKS. DO IT.
        LIST_FIELDS = [
            "IP_SET",
            "APPLIANCE_IDS",
            "DNS_LIST",
            "HOST_IDS",
            "NETBIOS_LIST",
            "ASSIGNED_USER_IDS",
            "ASSIGNED_UNIT_IDS",
            "EC2_IDS",
            "COMMENTS",
            "DOMAIN_LIST",
        ]

        # Iterate over the attrs of the dataclass and convert them to the appropriate format for SQL insertion.

        for attr in self.__dataclass_fields__.keys():
            if getattr(self, attr):
                if attr in LIST_FIELDS:
                    setattr(self, attr, str(getattr(self, attr)))

        sql_dict = self.to_dict()

        return sql_dict



"""
hosts.py - contains the dataclass for a Qualys GAV host record.
"""

from dataclasses import dataclass
from typing import List, Optional


@dataclass(frozen=True)
class Host:
    """
    Host - represents a Qualys GAV host record.

    due to the fact that some APIs have excludeFields and IncludeFields parameters,
    virtually all fields are optional other than assetId.
    """

    assetId: int = None
    assetUUID: Optional[str] = None
    hostId: Optional[int] = None
    lastModifiedDate: Optional[str] = None
    agentId: Optional[str] = None
    createdDate: Optional[str] = None
    sensorLastUpdatedDate: Optional[str] = None
    assetType: Optional[str] = None
    address: Optional[str] = None
    dnsName: Optional[str] = None
    assetName: Optional[str] = None
    netbiosName: Optional[str] = None
    timeZone: Optional[str] = None
    biosDescription: Optional[str] = None
    lastBoot: Optional[str] = None
    totalMemory: Optional[int] = None
    cpuCount: Optional[int] = None
    lastLoggedOnUser: Optional[str] = None
    domainRole: Optional[str] = None
    hwUUID: Optional[str] = None
    biosSerialNumber: Optional[str] = None
    biosAssetTag: Optional[str] = None
    isContainerHost: Optional[bool] = None
    operatingSystem: Optional[str] = None
    hardwareVendor: Optional[str] = None
    hardware: Optional[dict] = None
    userAccountListData: Optional[dict] = None
    openPortListData: Optional[dict] = None
    volumeListData: Optional[dict] = None
    networkInterfaceListData: Optional[dict] = None
    softwareListData: Optional[dict] = None
    softwareComponent: Optional[str] = None
    provider: Optional[str] = None
    cloudProvider: Optional[str] = None
    agent: Optional[dict] = None
    sensor: Optional[dict] = None
    container: Optional[dict] = None
    inventory: Optional[dict] = None
    activity: Optional[dict] = None
    tagList: Optional[List[str]] = None
    serviceList: Optional[List[str]] = None
    lastLocation: Optional[dict] = None
    criticality: Optional[int] = None
    businessInformation: Optional[dict] = None
    assignedLocation: Optional[dict] = None
    businessAppListData: Optional[dict] = None
    riskScore: Optional[int] = None
    passiveSensor: Optional[dict] = None
    domain: Optional[str] = None
    subdomain: Optional[str] = None
    missingSoftware: Optional[list] = None
    whois: Optional[dict] = None
    organizationName: Optional[str] = None
    isp: Optional[str] = None
    asn: Optional[str] = None
    easmTags: Optional[List[str]] = None
    hostingCategory1: Optional[str] = None
    customAttributes: Optional[dict] = None
    processor: Optional[dict] = None

    def is_cloud_host(self) -> bool:
        """
        Returns True if the host is a cloud host, False otherwise.
        """
        return self.cloudProvider is not None

    def is_container_host(self) -> bool:
        """
        Returns True if the host is a container host, False otherwise.
        """
        return self.container is not None

    def has_agent(self) -> bool:
        """
        Returns True if the host has an agent, False otherwise.
        """
        return self.agentId is not None

    def to_dict(self) -> dict:
        """
        Returns a dictionary representation of the host object.
        """
        return self.__dict__

    def keys(self) -> list:
        """
        Returns a list of keys for the host object.
        """
        return [key for key in self.__dict__.keys()]

    def values(self) -> list:
        """
        Returns a list of values for the host object.
        """
        return [value for value in self.__dict__.values()]

    def valid_values(self) -> list:
        """
        Return a list of keys that have values.
        """
        return [key for key, value in self.__dict__.items() if value is not None]

    def __iter__(self):
        """
        Allows for iteration over the host object.
        """
        for key, value in self.__dict__.items():
            yield key, value

    def __str__(self) -> str:
        """
        String representation of the host object.
        """
        return f"AssetID({self.assetId}))"

    def __int__(self) -> int:
        """
        Integer representation of the host object.
        """
        return self.assetId

    def __repr__(self) -> str:
        """
        Breaking the unwritten rule of having repr be something
        that can create the object for terminal space's sake.
        """
        return f"AssetID({self.assetId})"



"""
qualysdk - Qualys API SDK for Python

This package aims to make it easier to interact with the Qualys API across all of the different modules that Qualys provides.
"""

from .help import schema_query
from .auth import BasicAuth, TokenAuth

from .gav.uber import GAVUber

from .vmdr import query_kb, get_host_list, get_hld

from .sql import db_connect

# surprise!
__surprise__ = b"\xe2\x9c\xa8\xe2\x9c\xa8\xe2\x9c\xa8 Have a great day!".decode("utf-8")



"""
SQL server integration for qualysdk
"""

from .base import db_connect
from .supported_uploads import (
    upload_vmdr_ags,
    upload_vmdr_kb,
    upload_vmdr_hosts,
    upload_vmdr_ips,
    upload_vmdr_hld,
    upload_vmdr_scanners,
    upload_static_searchlists,
)



"""
help.py - Gives help on the different endpoints and what to expect from them.
"""

from json import dumps
from typing import Union

from .base.call_schema import CALL_SCHEMA


def schema_query(
    module: str, endpoint: str = None, pretty: bool = False
) -> Union[str, dict]:
    """
    Using the CALL_SCHEMA dictionary, this function will return the information for
    either a qualys module as a whole or a specific endpoint within a module.

    Params:
        module (str): The module to get help for.
        endpoint (str) [optional]: The endpoint to get help for.
        pretty (bool) [optional]: Whether to return the information as a string or as a dictionary. [False = return dict, True = return indented str]
    """

    module = module.lower()
    endpoint = endpoint.lower() if endpoint is not None else None

    # check if the module is in the CALL_SCHEMA:
    if module not in CALL_SCHEMA.keys():
        raise ValueError(
            f"Invalid module {module}. Available modules are: {CALL_SCHEMA.keys()}"
        )

    # if the endpoint is not provided, return the entire module schema:
    if endpoint is None:
        if pretty:
            return print(dumps(CALL_SCHEMA[module], indent=4))
        else:
            return CALL_SCHEMA[module]

    # if the endpoint is provided, return the schema for that endpoint:
    if endpoint not in CALL_SCHEMA[module].keys():
        raise ValueError(f"Invalid endpoint {endpoint} for module {module}.")

    if pretty:
        return print(dumps(CALL_SCHEMA[module][endpoint], indent=4))
    else:
        return CALL_SCHEMA[module][endpoint]



"""
base.py - contains the base functionality for the SQL module of qualysdk.
"""

from datetime import datetime
from dataclasses import dataclass
from typing import Literal
from ipaddress import IPv4Address, IPv6Address, IPv4Network, IPv6Network

from sqlalchemy import types

try:
    import pyodbc
except ImportError as e:
    raise Warning(
        "Pyodbc cannot be imported. If you are on Mac or Unix, this usually can be resolved by installing unixODBC:\n\nMac: brew install unixodbc\nUbuntu: sudo apt install unixodbc"
    ) from e

from pandas import DataFrame
from sqlalchemy import create_engine, Connection, types


def db_connect(
    host: str,
    db: str,
    username: str = None,
    password: str = None,
    driver: str = "ODBC Driver 17 for SQL Server",
    trusted_cnxn: bool = False,
    db_type: Literal["mssql", "mysql", "postgresql", "sqlite", "oracle"] = "mssql",
) -> Connection:
    """

    Generate a sqlalchemy Connection object to a SQL database.

    Parameters:

        host (str): The hostname of the SQL server.
        db (str): The name of the database.
        username (str): The username to use to connect to the database.
        password (str): The password to use to connect to the database.
        driver (str): The ODBC driver to use to connect to the database. Defaults to 'ODBC Driver 17 for SQL Server'.
        trusted_cnxn (bool): If True, use trusted connection. If False, use username and password. Defaults to False.
        db_type (str): The type of database to connect to. Defaults to 'mssql'. Options are 'mssql', 'mysql', 'postgresql', 'sqlite', 'oracle'.

    Returns:

        Connection: The Connection object to the SQL database.
    """

    # Check if user AND password are provided, or trusted connection is used:
    if not (username and password) and not trusted_cnxn:
        raise ValueError(
            "You must provide a username and password, or use trusted connection."
        )

    if driver not in pyodbc.drivers():
        if len(pyodbc.drivers()) == 0:
            raise ValueError(
                "No ODBC drivers found. Please install an ODBC driver & specify it with the driver parameter."
            )
        else:
            raise ValueError(
                f"Driver '{driver}' not found. Pyodbc found drivers: {pyodbc.drivers()}"
            )

    # Generate the connection string:
    if trusted_cnxn:
        conn_str = (
            f"{db_type}+pyodbc://{host}/{db}?driver={driver}&trusted_connection=yes"
        )
    else:
        conn_str = (
            f"{db_type}+pyodbc://{username}:{password}@{host}/{db}?driver={driver}"
        )

    engine = create_engine(conn_str)

    return engine.connect()


def upload_data(
    df: DataFrame,
    table: str,
    cnxn: Connection,
    dtype: dict,
    override_import_dt: datetime = None,
) -> int:
    """
    Upload a DataFrame to a SQL table. Appends 'import_datetime' column to the DataFrame.

    Parameters:
        df (DataFrame): The DataFrame to upload.
        table (str): The name of the table to upload to.
        cnxn (Connection): The Connection object to the SQL database.
        dtype (dict): The data types of the columns in the table. Key is the column name, value is the data type as sqlalchemy.types.Something()
        override_import_dt (datetime): If provided, will override the import_datetime column with this value.

    Returns:
    int: The number of rows uploaded.
    """

    # Add an import_datetime column:
    df["import_datetime"] = (
        datetime.now() if not override_import_dt else override_import_dt
    )
    dtype["import_datetime"] = types.DateTime()

    # Upload the data:
    print(f"Uploading {len(df)} rows to {table}...")
    df.to_sql(table, cnxn, if_exists="append", index=False, dtype=dtype, method=None)

    return len(df)


def prepare_dataclass(dataclass: dataclass) -> dict:
    """
    Prepare the dataclass for insertion into a SQL database
    by converting it to a dictionary with appropriate list/dataclass
    conversions.

    Parameters:
    dataclass (dataclass): The dataclass to convert.

    Returns:
    dict: The dataclass converted to a dictionary.
    """

    TO_STR_FIELDS = [
        "IP_SET",
        "APPLIANCE_IDS",
        "DNS_LIST",
        "HOST_IDS",
        "NETBIOS_LIST",
        "ASSIGNED_USER_IDS",
        "ASSIGNED_UNIT_IDS",
        "EC2_IDS",
        "COMMENTS",
        "DOMAIN_LIST",
        "BUGTRAQ_LIST",
        "SOFTWARE_LIST",
        "VENDOR_REFERENCE_LIST",
        "CVE_LIST",
        "THREAT_INTELLIGENCE",
        "COMPLIANCE_LIST",
        "TAGS",
        "CLOUD_PROVIDER_TAGS",
        "IP",
        "IPV6",
        "QDS",
        "QDS_FACTORS",
        "QIDS",
    ]

    DICT_FIELDS = [
        "CORRELATION",
        "CVSS",
        "CVSS_V3",
        "PCI_REASONS",
        "DISCOVERY",
        "CHANGE_LOG",
        "USER_DEF",
        "TRURISK_SCORE_FACTORS",
        "VLANS",
        "ML_VERSION",
        "VULNSIGS_VERSION",
        "OPTION_PROFILES",
        "REPORT_TEMPLATES",
        "REMEDIATION_POLICIES",
        "DISTRIBUTION_GROUPS",
    ]

    # Iterate over the attrs of the dataclass and convert them to the appropriate format for SQL insertion.

    for attr in dataclass.__dataclass_fields__.keys():
        if getattr(dataclass, attr):
            if attr in TO_STR_FIELDS:
                setattr(dataclass, attr, str(getattr(dataclass, attr)))
            elif attr in DICT_FIELDS:
                setattr(
                    dataclass, attr, flatten_dict_to_string(getattr(dataclass, attr))
                )

    # If there are any leftover empties or ipv4/ipv6 addresses,
    # convert them to None/str as a failsafe: #NOTE: Refactor at some point!
    for attr in dataclass.__dataclass_fields__.keys():
        """
        if getattr(dataclass, attr) in {}:
            setattr(dataclass, attr, None)
        elif getattr(dataclass, attr) == "":
            setattr(dataclass, attr, None)
        elif getattr(dataclass, attr) == []:
            setattr(dataclass, attr, None)
        """
        if not getattr(dataclass, attr):
            setattr(dataclass, attr, None)
        elif type(getattr(dataclass, attr)) not in [str, int, float, bool, datetime]:
            setattr(dataclass, attr, str(getattr(dataclass, attr)))
        """elif type(getattr(dataclass, attr)) in [
            IPv4Address,
            IPv6Address,
            IPv4Network,
            IPv6Network,
        ]:
            setattr(dataclass, attr, str(getattr(dataclass, attr)))
        """
    sql_dict = dataclass.to_dict()

    return sql_dict


def flatten_dict_to_string(d, parent_key="") -> str:
    """
    Format dictionary fields to a string for SQL insertion.
    """
    items = []
    for k, v in d.items():
        new_key = f"{parent_key}{k}" if parent_key else k
        if isinstance(v, dict):
            items.append(flatten_dict_to_string(v, new_key))
        else:
            items.append(f"{new_key}:{v}")
    return ", ".join(items)



"""
compliance.py - contains the Compliance dataclass for the Qualys VMDR module.
"""

from dataclasses import dataclass, field, asdict
from typing import *


@dataclass
class Compliance:
    """
    Compliance - represents a compliance object.
    """

    _TYPE: str = field(metadata={"description": "The compliance framework name."})
    SECTION: str = field(metadata={"description": "The section name."})
    DESCRIPTION: str = field(
        metadata={"description": "The description of the compliance."}
    )

    def __str__(self):
        return f"{self._TYPE} - {self.SECTION}"

    def __contains__(self, item):
        return item in self._TYPE or item in self.SECTION or item in self.DESCRIPTION

    def copy(self):
        return Compliance(
            _TYPE=self._TYPE, SECTION=self.SECTION, DESCRIPTION=self.DESCRIPTION
        )

    def is_type(self, _type: str):
        return self._TYPE == _type

    def is_section(self, section: str):
        return self.SECTION == section

    def is_description(self, description: str):
        return self.DESCRIPTION == description

    def __dict__(self):
        return asdict(self)

    def to_dict(self):
        return asdict(self)

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    @classmethod
    def from_dict(cls, data: dict):
        """
        from_dict - create a Compliance object from a dictionary.

        Params:
            data (dict): The dictionary containing the data for the Compliance object.

        Returns:
            Compliance: The Compliance object created from the dictionary.
        """
        required_keys = {"_TYPE", "SECTION", "DESCRIPTION"}
        if not required_keys.issubset(data.keys()):
            raise ValueError(
                f"Dictionary must contain the following keys: {required_keys}"
            )
        return cls(**data)



"""
qds_factor.py - contains the QDSFactor dataclass for the Qualys VMDR module.
"""

from dataclasses import dataclass, field, asdict
from typing import *


@dataclass
class QDSFactor:
    """
    QDSFactor - represents a single Qualys Detection Score (QDS) factor.
    """

    NAME: int = field(metadata={"description": "The name of the factor."})
    VALUE: Union[str, int, float] = field(
        metadata={
            "description": "The value of the factor. Can be a str or number depending on what the @name is."
        },
        default="",
    )

    def __post_init__(self):
        # make sure that the ID is an integer:
        if not isinstance(self.NAME, str):
            raise TypeError(f"QDSFactor NAME must be a string, not {type(self.NAME)}")
        # check if the VALUE can be interpreted as a number:
        for val in (float, int):
            try:
                # special case for CVSS version, as qualys prepends as 'v' to the number:
                if self.NAME == "CVSS":
                    self.VALUE = self.VALUE[1:]

                self.VALUE = val(self.VALUE)
                break
            except ValueError:
                pass

    def __str__(self):
        return str(self.VALUE)

    def __contains__(self, item):
        return item in self.NAME or item in self.VALUE

    def copy(self):
        return QDSFactor(ID=self.NAME, TEXT=self.VALUE)

    def is_id(self, id: int):
        return self.NAME == id

    def is_text(self, text: str):
        return self.VALUE == text

    def has_rti(self):
        return self.NAME == "RTI"

    def has_cvss_score(self):
        return self.NAME == "CVSS"

    def has_epss(self):
        return self.NAME == "epss"

    def get_epss(self):
        if self.has_epss():
            return self.VALUE

    def get_cvss_score(self):
        if self.has_cvss_score():
            return self.VALUE

    def has_malware_hash(self):
        return self.NAME == "malware_hash"

    def get_malware_hash(self):
        if self.has_malware_hash():
            return self.VALUE

    def get_rti(self):
        if self.has_rti():
            return self.VALUE

    def get_name(self):
        return self.NAME

    def to_dict(self):
        return asdict(self)

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    @classmethod
    def from_dict(cls, data: dict):
        """
        from_dict - create a QDSFactor object from a dictionary.

        Params:
            data (dict): The dictionary containing the data for the QDSFactor object.

        Returns:
            QDSFactor: The QDSFactor object created from the dictionary.
        """
        required_keys = {"NAME", "VALUE"}
        if not required_keys.issubset(data.keys()):
            raise ValueError(
                f"Dictionary must contain the following keys: {required_keys}"
            )
        return cls(**data)
